/*******************************************************************
*  Copyright (c) 1994-2008 Jetico, Inc., Finland
*  All rights reserved.
*
*  File:          wipe.c
*
*  Description:   implementation of BestCrypt wipe utility
*
*  Author:        Vitaliy Zolotarev
*
*  Created:       20-Nov-1999
*
*  Revision:      $Id: wipe.c 282 2009-09-09 11:44:43Z nail $
*
*
*******************************************************************/
#include "config.h"
#include <stdio.h>
#include <limits.h>
#include <stdlib.h>
#include <sys/types.h>
#include <stdarg.h>
#include <dirent.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <string.h>
#include <utime.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <time.h>
#include <ctype.h>

#if HAVE_SYS_DISKIO_H /* HP-UX */
#include <sys/diskio.h>
#endif

#if HAVE_SYS_STATVFS_H
#include <sys/statvfs.h>
#endif

#if HAVE_SYS_DISK_H
#include <sys/disk.h>
#endif

#ifndef NAME_MAX
#if defined(MAXNAMELEN)
#define NAME_MAX MAXNAMELEN
#elif defined(FILENAME_MAX) 
#define NAME_MAX FILENAME_MAX
#endif 
#endif 

#ifndef PATH_MAX
#ifdef MAXPATHLEN
#define PATH_MAX MAXPATHLEN
#endif 
#endif 

#ifndef S_BLKSIZE
#define S_BLKSIZE   512
#endif    

/*
// The statvfs is changed to statvfs_ because Solaris 2.8 reports:
// wipe.c:1676: warning: passing are 2 of \'statvfs64\' from incompatible pointer type	
*/
/*

#if (OS_NetBSD || OS_OpenBSD || OS_FreeBSD || OS_Darwin)

#define statvfs_ statfs
#define f_frsize f_bsize

#else

#if (__USE_LARGEFILE64 || __USE_FILE_OFFSET64)
#define statvfs_f statvfs64
#else
#define statvfs_f statvfs
#endif
#endif
*/

#include "wipe.h"
#include "options.h"
#include "schemes.h"
#include "sha1random.h"
#include "log.h"

typedef time_t	seed_t;

typedef struct wipe_info 
{
	long long int	 size;		/* actual size, may be 0 for devices */
	long long int	 slack_size;	/* file slack size */
	char		 *filename;	/* name */
	mode_t		 st_mode;	/* st_mode from stat()/lstat() call */
	int		 pass;		/* current pass */
	seed_t		 seed;		/* seed to initialize pseudo-random sequence */
	seed_t		 prev_seed;	/* seed for previous pass */
	int		 skip;		/* an error occured, error number here */
	struct wipe_info *next;
} wipe_info_t;


static void                   *sRandom;
static unsigned long long int total_size_of_slacks=0;
static wipe_info_t            *FileListHead = NULL;
static wipe_scheme_t          scheme;


/* ---------------------------------------------------------------------------------------------- */


long long int get_size_of_device( int fd );

int wipe_file(char *filename);
int wipe_filename(char *filename, int delete_file_contents);
int wipe_dir(char *dirname);
int wipe_dirname(char *dirname);

int nas_wiping(int argc, char *argv[]);

int add_file_to_list(char *FileName);
int add_dir_to_list(char *DirName);

int remove_from_list(wipe_info_t *ptr);
int free_list(wipe_info_t *head);

int wipe_files_in_list( int pass, int verify );
int wipe_filenames_in_list();
int wipe_file_nas( wipe_info_t *ctx, int pass, int verify );
int wipe_file_slack_nas( wipe_info_t *ctx, int pass, int verify );


/* SHA-1 random number generator helpers */

static int StartRandom() 
{
	time_t t;
	int    result;

	t=time(&t);
	srand(t);

	result = SHA1RandomAllocate( &sRandom );
	if ( result != 0 ) {
		log_message(LOG_FATAL, "Could not allocate memory for SHA1 random generator\n");
		return result;
	}

	result = SHA1RandomInitialize( sRandom );
	if ( result != 0 ) {
		log_message(LOG_FATAL, "Could not initialize SHA1 random generator\n");
		return result;
	}

	result = SHA1RandomRandomize( sRandom, (byte*)&t, sizeof(t));
	if ( result != 0 ) {
		log_message(LOG_FATAL, "Could not randomize SHA1 random generator\n");
		return result;
	}
	
	return result;
}

static void StopRandom() 
{
	SHA1RandomFinalize( sRandom );
	SHA1RandomFree( &sRandom );
}

static void RestartRandom( time_t t )
{
    srand( t );
		
    SHA1RandomFinalize( sRandom );
    SHA1RandomInitialize( sRandom );
    SHA1RandomRandomize( sRandom, (byte*)&t, sizeof(t));
}

/*****************************************/
static ssize_t bcwipe_read(int fd, void *buf, size_t count) 
{
	ssize_t	rc, n;
	off_t	offset;
	
	n = 0;
	offset = lseek_f(fd, 0, SEEK_CUR);
	do {
		rc = read(fd, buf+n, count-n);
		if (-1 == rc) {
			if (EINTR == errno || EAGAIN == errno || EWOULDBLOCK == errno) {
				lseek_f(fd, offset+n, SEEK_SET);
				continue;
			}
			return -1;
		}
		n += rc; 
		if (0 == rc) {
		    return n;
		}
	} while (count > n);
	
	return n;
}

static ssize_t bcwipe_write(int fd, void *buf, size_t count) 
{
#ifndef SAFE_BCWIPE_WRITE
	return write(fd, buf, count);
#else
	ssize_t	rc, n;
	off_t	offset;
	
	n = 0;
	offset = lseek_f(fd, 0, SEEK_CUR);
	do {
		rc = write(fd, buf+n, count-n);
		if (-1 == rc) {
			if (EINTR == errno || EAGAIN == errno || EWOULDBLOCK == errno) {
				lseek_f(fd, offset+n, SEEK_SET);
				continue;
			}
			return -1;
		}
		n += rc; 
		if (0 == rc) {
		    return n;
		}
	} while (count > n);
	return n;
#endif
}

static ssize_t verify_buffer(char *buf, char *pattern, ssize_t len) 
{
	ssize_t i;

	if (0 == memcmp(buf, pattern, len)) {
		return 0;
	}
	
	for (i = 0; i < len; i++) {
		if (buf[i] != pattern[i])
			break;
	
	}
	return i+1;
}

int ask_y_n(char *message, ... )
{
	va_list var;
	char	c;
	
	va_start(var, message);
	vfprintf(stderr, message, var);
	va_end(var);

	c = tolower(fgetc(stdin));
	if ( '\n' == c ) return FALSE;
	while ( '\n' != fgetc(stdin) );
	if ( 'y' == c ) return TRUE;
	if ( 'a' == c )
	{
		o_interactive=FALSE;
		return TRUE;
	}

	return FALSE;
}

static void print_wiping_progress(long long int total_size, long long int file_size, time_t start_time, int pass) 
{
	time_t cur_time;
	int    percent;
	
	cur_time = time(NULL);
	percent = (int)(file_size * 100 / total_size);

	fprintf( stdout, "\rwipe pass %2d/%-2d:%10ld/%ld kB (%3d%%)", 
	pass+1, o_pas_num, (long int)(file_size >> 10), (long int)(total_size >> 10), percent );

	if ( cur_time != start_time ) {
		fprintf(stdout,"   Rate: %ld kB/s          ", (long int)((file_size/(long long int)(cur_time - start_time)) >> 10));
	} else	{
		fprintf(stdout,"                        ");
	}
}

static void print_verification_progress(long long int total_size, long long int file_size, time_t start_time, int pass)
{
	time_t cur_time;
	int    percent;

	cur_time = time(NULL);
	percent =(int)(file_size * 100 / total_size);
			
	fprintf( stdout, "\rverify pass %-2d :%10ld/%ld kB (%3d%%)",pass+1, (long int)(file_size >> 10),(long int)(total_size >> 10), percent );
	if ( start_time != cur_time ) {
		fprintf(stdout, "   Rate: %ld kB/s          ",
		(long int)((file_size /(long long int)(cur_time - start_time)) >> 10));
	} else {
		fprintf(stdout, "                                ");
	}
}

static char *get_type_str(mode_t st_mode) 
{
	if ( S_ISDIR(st_mode) ) {
		return "directory";
	} else if ( S_ISREG(st_mode) ) {
		return "file";
	} else if ( S_ISLNK(st_mode) ) {
		return "link";
	} else if ( S_ISCHR(st_mode) ) {
		return "char device";
	} else if ( S_ISBLK(st_mode) ) {
		return "block device";
	} else if ( S_ISFIFO(st_mode) ) {
		return "pipe";
	} else if ( S_ISSOCK(st_mode) ) {
		return "socket";
	}
	return "";
}

static int is_device(mode_t st_mode) 
{
	/*
	//	v. 1.7-2 Solaris does not recognize link as block device.
	//	So we try to force block device reading if o_wipe_dev flag is assigned
	*/
	return ( S_ISBLK(st_mode) || S_ISCHR(st_mode) || (o_wipe_dev && S_ISLNK(st_mode)) );
}

/*
	fill_buff - prepare buffer for wiping/verification
	
	return values:	1 - buffer overwritten with new values
			0 - buffer kept intact
			-1 - an error occured
*/
int fill_buff(char *buff, size_t size, int pass, off_t pos, seed_t seed, seed_t prev_seed ) 
{
	int i, j, len;
	if (pass > scheme.num_passes) {
		return -1;
	}

	len = scheme.pass[pass].len;
	if (len) {
		if (!pos) {
			for (i = 0, j = 0; i < size; i++)
				buff[i] = scheme.pass[pass].pat[j++ % len];
		}
		return 1;
	}
	
	switch (scheme.pass[pass].type) {
	case PASS_RANDOM:
		if (!pos) {
			RestartRandom(seed);
		}
		if (pos && o_use_buff) {
			return 0;
		}
		if (o_use_rand) {
			for (i = 0; i < size; i++) 
				buff[i] = (char)(rand() & 0xff);
		} else {
			SHA1RandomGetRandomBytes(sRandom, (byte*)buff, size);
		}
		return 1;

	case PASS_COMPLEMENT:
		if (0 == pass) {
			memset(buff, 0, size);
			return 1;
		}
		i = fill_buff(buff, size, pass-1, pos, prev_seed, 0);
		if (1 > i) {
			return -1;
		}
		for (i = 0; i < size; i++) {
			buff[i]=~buff[i];
		}
		return 1;

	case PASS_ZERO:
		if (!pos) {
			memset(buff, 0, size);
		}
		return 1;
	
	case PASS_TEST:
		memset( buff, 0, size );
		for (i = 0; i < size; i += 512) {
			j = (pos+i)/512;
			buff[i+0] = (j >> 24) & 0xff;
			buff[i+1] = (j >> 16) & 0xff;
			buff[i+2] = (j >> 8)  & 0xff;
			buff[i+3] =  j        & 0xff;
		}
		return 1;
	default:
		return -1;
	}
}

/*****************************************/
/*
static void showBuffer( char *bfName, char *bf, int size )
{
    int col,i;

    printf( "Buffer %s: \n", bfName );
    
    for( col = 0, i = 0; i < size; i++, col++ )
    {
	if ( col >= 16 )
	{
	    printf( "\n" );
	    col = 0;
	}
	
	if ( col == 0 )
	    printf( "%08d ", i );
	
	unsigned int ch = (unsigned int)bf[i];
	printf( "%X%X ", (ch>>4)&0xF, (ch)&0xF );
    }
    
    printf("\n");
}


int print_list(WipeFileEntry *ListHead)
{
    while (NULL != ListHead) {
	printf("%s\ttype=%d\tsize=%Ld\tslack=%Ld\n",
		ListHead->FileName,
		ListHead->Type,
		ListHead->FileSize,
		ListHead->SlackSize);
	ListHead=ListHead->Next;
    }
    return OK;
}
*/

int nas_wiping(int argc, char *argv[])
{
	int i, pass, save;

	log_message(LOG_INFO, "Wiping in NAS mode\n");

	/* fill wiping list */
	for ( i = argc; i; i-- ) {
		add_file_to_list( argv[i-1] );
	}
	if (NULL == FileListHead) {
		log_message(LOG_ERR, "Could not build list of files for wiping\n");
		return ERROR;
	}

	save = o_dont_delete;
	o_dont_delete = TRUE;

	for (pass = 0; pass < o_pas_num; pass++) {
		wipe_files_in_list(pass, FALSE);
		log_message( LOG_INFO, "Sleeping %d seconds\n", o_nas_delay );
		sleep(o_nas_delay);
	}
	
	if (o_verify_last_pass) {
		wipe_files_in_list(pass ? pass-1 : 0, TRUE );
	}

	o_dont_delete = save;

	if ( !o_wipe_dev ) {
		wipe_filenames_in_list();
	}
	
	free_list(FileListHead);
	FileListHead = NULL;

	return OK;
}


int add_file_to_list(char *filename)
{
	struct stat buff;
	wipe_info_t *ptr = NULL;

	if (!filename)	
		return ERROR;

	if ( OK != lstat(filename, &buff) ) {
		log_message (LOG_ERR, "Skipping '%s': %s\n", filename, strerror(errno));
		return ERROR;
	}

	if ( is_device(buff.st_mode) ) {
		/* Get stat instead of lstat. */
		if ( ERROR == stat( filename, &buff ) )	{
			log_message (LOG_ERR, "Skipping '%s': %s\n", filename, strerror(errno));
			return ERROR;
		}
		
		if ( !S_ISBLK( buff.st_mode ) && !S_ISCHR( buff.st_mode ) ) {
			log_message(LOG_ERR, "'%s': is not a special device - skipping\n", filename );
			return ERROR;	
		}
	}

	ptr = (wipe_info_t *)malloc( sizeof( wipe_info_t ) );
	if (!ptr) {
		log_message(LOG_ERR, "Memory allocation error. Skipping '%s'\n", filename, strerror(errno));
		return ERROR;
	}
	memset( ptr, 0, sizeof(wipe_info_t) );

	if ( S_ISDIR(buff.st_mode) ) {
		if (OK != add_dir_to_list( filename ) ) {
			free(ptr);
			return ERROR;
		}
	} else if ( S_ISREG(buff.st_mode) ) {
		ptr->size  = buff.st_size;
		ptr->slack_size = buff.st_size % buff.st_blksize ? buff.st_blksize - (buff.st_size % buff.st_blksize) : 0;
	} else if ( S_ISLNK(buff.st_mode)  || 
		    S_ISCHR(buff.st_mode)  ||
		    S_ISBLK(buff.st_mode)  ||
		    S_ISFIFO(buff.st_mode) ||
		    S_ISSOCK(buff.st_mode) ) {
		;
	} else { 
		log_message(LOG_ERR, "Unsupported type of file: '%s'\n", filename);
		free(ptr);
		return ERROR;
	}

	if ( o_interactive && 
	     !ask_y_n("Add %s '%s' to wiping list (y/[n]/a)?", get_type_str(buff.st_mode), filename) ) {
		free(ptr);
		return ERROR;
	}

	ptr->st_mode = buff.st_mode;
	ptr->filename = strdup(filename);
	if (NULL == ptr->filename) {
		log_message(LOG_ERR, "Skipping '%s': %s\n", filename, strerror(errno));
		free(ptr);
		return ERROR;
	}

	ptr->next = FileListHead;
	FileListHead = ptr;

	return OK;
}

int add_dir_to_list(char *dirname)
{
	struct dirent *d;
	DIR *fd;
	char filename[NAME_MAX+PATH_MAX];

	if ( !dirname ) 
		return ERROR;

	if ( !o_recurse ) {
		log_message(LOG_INFO, "'%s' is a directory - skipping\n", dirname);
		return ERROR;
	}

	/* check directory access */
	if ( !o_wipe_slacks && OK != access( dirname, X_OK|R_OK|W_OK ) ) {
		if ( !o_force ) {
			log_message(LOG_ERR, "Could not enter to directory '%s': %s - skipping\n", dirname, strerror(errno));
			return ERROR;
		}
		if ( OK != chmod(dirname, S_IRUSR|S_IWUSR|S_IXUSR ) ) {
			log_message(LOG_ERR, "Could not change permissions for directory '%s': %s - skipping\n", dirname, strerror(errno));
			return ERROR;
		}
	}

	fd = opendir(dirname);
	if ( NULL == fd ) {
		log_message(LOG_ERR, "Could not open directory '%s': %s - skipping\n", dirname, strerror(errno));
		return ERROR;
	}


	while ( NULL != (d = readdir( fd )) ) {
		snprintf(filename, NAME_MAX+PATH_MAX, "%s/%s", dirname, d->d_name);
		if ( 0 == strcmp(d->d_name,".") || 0 == strcmp(d->d_name,"..") )
			continue;
		add_file_to_list(filename);
	}

	closedir(fd);

	return OK;

}

int free_list(wipe_info_t *head)
{
	wipe_info_t *ptr;

	while (NULL != head) {
		ptr  = head;
		head = ptr->next;
		free(ptr->filename);
		free(ptr);
	}
	return OK;
}


int wipe_files_in_list( int pass, int verify )
{
	wipe_info_t *ptr = FileListHead;
	int          rc;

	while ( NULL != ptr ) {
		rc = wipe_file_nas( ptr, pass, verify );
		if ( OK != rc ) {
			ptr->skip = rc;
		}
		ptr = ptr->next;
	}
	return OK;
}

static int do_open_file(wipe_info_t *ctx) 
{
	int fd;
	
	if ( OK != access(ctx->filename, W_OK|R_OK) ) {
		if ( !o_force ) {
			log_message(LOG_ERR, "Access to %s '%s' denied: %s - skipping\n", get_type_str(ctx->st_mode), ctx->filename, strerror(errno));
			return ERROR;
		}
		
		if ( OK != chmod(ctx->filename, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH) ) {
			log_message(LOG_ERR, "Could not change '%s' permissions: %s - skipping\n", ctx->filename, strerror(errno));
			return ERROR;
		}
	}

	fd = open(ctx->filename, O_RDWR);
	if ( ERROR == fd ) {
		log_message(LOG_ERR, "Could not open %s '%s': %s - skipping\n", get_type_str(ctx->st_mode), ctx->filename, strerror(errno));
		return ERROR;
	}
	return fd;
}

static int do_open_file_slack(wipe_info_t *ctx) 
{
	int fd;
	
	if ( !S_ISREG(ctx->st_mode) ) {
		log_message(LOG_INFO, "'%s' is not a regular file\n", ctx->filename);
		ctx->skip = 1;
		return OK;
	}

	if (0 == ctx->slack_size) {
		log_message(LOG_INFO, "'%s' does not have a slack\n", ctx->filename);
		ctx->skip = 1;
		return OK;
	}

	if ( OK != access(ctx->filename, W_OK|R_OK) ) {
		if ( !o_force ) {
			log_message(LOG_ERR, "Access to %s '%s' denied: %s - skipping\n", get_type_str(ctx->st_mode), ctx->filename, strerror(errno));
			return ERROR;
		}
		
		if ( OK != chmod(ctx->filename, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH) ) {
			log_message(LOG_ERR, "Could not change '%s' permissions: %s - skipping\n", ctx->filename, strerror(errno));
			return ERROR;
		}
	}
/*
#ifdef O_SYNC
	fd = open(ctx->filename, O_RDWR|O_SYNC);
#else
	fd = open(ctx->filename, O_RDWR);
#endif
*/
	fd = open(ctx->filename, O_RDWR);
	if ( ERROR == fd ) {
		log_message(LOG_ERR, "Could not open file '%s': %s - skipping\n", ctx->filename, strerror(errno));
		return ERROR;
	}
	
	return fd;
}

static int do_wipe_pass(int fd, wipe_info_t *ctx) 
{
	long long int	pos, len, size;
	time_t		start_time;
	char		buff[BUFFSIZE];

	print_pass_name(buff, BUFFSIZE, &scheme, ctx->pass);
	log_message(LOG_INFO, "Wiping %s '%s' pass %d/%d [%s] started\n", get_type_str(ctx->st_mode), ctx->filename, ctx->pass+1, o_pas_num, buff);

	lseek_f(fd, 0, SEEK_SET);

	size = ctx->size + ctx->slack_size;
	pos = 0;
	start_time = time(NULL);

	while ( pos < size ) {
		len = size - pos < BUFFSIZE ? size - pos : BUFFSIZE;
		fill_buff( buff, len, ctx->pass, pos, ctx->seed, ctx->prev_seed );

		len = bcwipe_write(fd, (void *)buff, len);
		if ( ERROR == len ) {
			if ( ( ENOSPC == errno ) && is_device(ctx->st_mode) ) {
 				ctx->size = pos;
				len = 0;
			} else {
				log_message(LOG_ERR, "Writing to %s '%s' failed at %lld bytes: %s\n", get_type_str(ctx->st_mode), ctx->filename, pos, strerror(errno));
				goto error_out;;
			}
		}

		pos += len;

		if ( o_verbose ) {
			print_wiping_progress(size, pos, start_time, ctx->pass);
		}
		
		if (len != BUFFSIZE) {
			break;
		}
	}

	fsync(fd);

	if ( o_verbose ) {
		fprintf( stdout, "\n");
	}
	
	log_message(LOG_INFO, "Wiping %s '%s' pass %d/%d completed\n", get_type_str(ctx->st_mode), ctx->filename, ctx->pass+1, o_pas_num);
	return OK;

error_out:

	log_message(LOG_ERR, "Wiping %s '%s' pass %d/%d failed\n", get_type_str(ctx->st_mode), ctx->filename, ctx->pass+1, o_pas_num);
	return ERROR;
}


static int do_verify_pass(int fd, wipe_info_t *ctx) 
{
	long long int pos, len, size;
	time_t	cur_time, start_time;
	int x;
	char verifyBuff[BUFFSIZE];
	char buff[BUFFSIZE];

	log_message(LOG_INFO, "Wiping %s '%s' pass %d verification started\n", get_type_str(ctx->st_mode), ctx->filename, ctx->pass+1);
	
	lseek_f(fd, 0, SEEK_SET);

	size = ctx->size;
	pos = 0;
	start_time = time(NULL);

	while ( pos < size ) {
		len = size - pos < BUFFSIZE ? size - pos : BUFFSIZE;
		fill_buff( buff, len, ctx->pass, pos, ctx->seed, ctx->prev_seed );

		len = bcwipe_read(fd, (void *)verifyBuff, len);
		if ( ERROR == len )  {
			log_message(LOG_ERR, "Error reading '%s' at %lld bytes : %s\n", ctx->filename, pos, strerror(errno) );
			goto error_out;
		}
		x = verify_buffer( verifyBuff, buff, len );
		if ( 0 != x ) {
			log_message(LOG_ERR, "Verification of %s '%s' failed at %lld bytes\n"
					     "Expected %02Xh, read %02Xh\n", 
				    get_type_str(ctx->st_mode), ctx->filename, pos + x - 1, (unsigned char)verifyBuff[x-1], (unsigned char)buff[x-1] );
			goto error_out;
		}

		pos += len;
		if ( o_verbose ) {
			print_verification_progress(size, pos, start_time, ctx->pass);
		}
		
		if (len != BUFFSIZE) {
			break;
		}
	}
	if ( o_verbose ) {
		fprintf( stdout, "\n");
	}

	log_message(LOG_INFO, "Wiping %s '%s' pass %d verification completed\n", get_type_str(ctx->st_mode), ctx->filename, ctx->pass+1);
	return OK;
	
error_out:

	if ( o_verbose ) {
		fprintf( stdout, "\n");
	}
	log_message(LOG_ERR, "Wiping %s '%s' pass %d verification failed\n", get_type_str(ctx->st_mode), ctx->filename, ctx->pass+1);
	return ERROR;
	
}

/*
// Verification process requires same random byte sequence
// The function inites all random generators to start from same initial value
// here a time varable is used.
*/

int wipe_file_nas( wipe_info_t *ctx, int pass, int verify )
{
	int fd, is_dev, ret;
	long long int size;
	char *type_str;

	if (o_wipe_slacks)
		return wipe_file_slack_nas( ctx, pass, verify );

	if (ctx->skip)
		return OK;

	if (S_ISDIR(ctx->st_mode))
		return OK;

	is_dev = is_device( ctx->st_mode );
	if ( 0 == ctx->size && !is_dev ) {
		if (0 == pass) {
			log_message(LOG_INFO, "Skipping %s '%s': nothing to do\n", get_type_str(ctx->st_mode), ctx->filename);
		}
		return OK;
	}

	fd = do_open_file(ctx);
	if (ERROR == fd) {
		return ERROR;
	}

	if ( is_dev && 0 == ctx->size ) {
		ctx->size = get_size_of_device(fd);
	}

	size = ctx->size + ctx->slack_size;

	if (verify) {
		pass = ctx->pass; /*  In verification process the last saved settings will be used. */
		do_verify_pass(fd, ctx);
	} else {
		ctx->prev_seed = ctx->seed;
		ctx->seed = time( NULL );
		ctx->pass = pass;
		do_wipe_pass(fd, ctx);
	}
	
	if ( !o_dont_delete ) {
		ret = ftruncate( fd, 0 );
	}
	close(fd);

	return OK;

}

int wipe_filenames_in_list()
{
	wipe_info_t *ptr = FileListHead;

	while (ptr) {
		if (!ptr->skip) {
			if (S_ISDIR(ptr->st_mode)) {
				wipe_dirname(ptr->filename);
			} else {
				wipe_filename(ptr->filename, S_ISREG(ptr->st_mode));
			}
		}
		ptr = ptr->next;
	}
	return OK;
}

int wipe_file_slack_nas(wipe_info_t *ctx, int pass, int verify)
{
	char *buff, msg[1024];
	int   fd, ret;
	struct stat st;
	struct utimbuf time_buf;

	if (ctx->skip)
		return OK;
		
	if (verify)
		return OK;

	if ( ERROR == lstat(ctx->filename, &st) ) {
		log_message(LOG_ERR, "Could not get information on '%s': %s - skipping\n", ctx->filename, strerror(errno));
		return ERROR;
	}

	fd = do_open_file_slack(ctx);
	if (0 == fd) {
		return OK;
	} else if (ERROR == fd) {
		return ERROR;
	}

	buff = (char *)malloc(ctx->slack_size);
	if (NULL == buff) {
		close(fd);
		log_message(LOG_ERR, "Memory allocation error: %s - skipping\n", strerror(errno));
		return ERROR;
	}

	print_pass_name(msg, 1024, &scheme, pass);
	log_message(LOG_INFO, "Wiping slack of '%s' pass %d/%d [%s] started\n", ctx->filename, pass+1, o_pas_num, msg);

	if (ERROR == lseek_f(fd, ctx->size, SEEK_SET)) {
		log_message(LOG_ERR, "Seek to end of file '%s': %s - skipping\n", ctx->filename, strerror(errno));
		ret = ftruncate(fd, ctx->size);
		close(fd);
		free(buff);
		return ERROR;
	}

	ctx->prev_seed = ctx->seed;
	ctx->seed = time(NULL);
	fill_buff(buff, ctx->slack_size, pass, 0, ctx->seed, ctx->prev_seed);
	
	if (ERROR == bcwipe_write(fd, (void *)buff, ctx->slack_size))	{
		log_message(LOG_ERR, "Error writing to '%s': %s\n", ctx->filename, strerror(errno));
		ret = ftruncate(fd, ctx->size);
		close(fd);
		free(buff);
		return ERROR;
	}

/*
#ifndef O_SYNC
	(void)fsync(fd);
#endif
*/
	(void)fsync(fd);
	log_message (LOG_INFO, "Wiping slack of '%s' pass %2d/%d completed\n", ctx->filename, pass+1, o_pas_num);

	if (o_pas_num == pass+1) {
		total_size_of_slacks += ctx->slack_size;
	}
	
	ret = ftruncate(fd, ctx->size);
	close(fd);
	free(buff);

	if ( OK != chmod(ctx->filename, st.st_mode) ) {
		log_message(LOG_ERR, "Restore access mask for file '%s': %s\n", ctx->filename, strerror(errno));
		return ERROR;
	}

	time_buf.actime  = st.st_atime; 
	time_buf.modtime = st.st_mtime;
	if ( OK != utime(ctx->filename, &time_buf) ) {
		log_message(LOG_ERR, "Restore access times of file '%s': %s\n", ctx->filename, strerror(errno));
		return ERROR;
	}

	return OK;
}

int wipe_dirname(char *dirname)
{
	struct utimbuf utb;

	if ( o_dont_delete || o_wipe_slacks) 
		return OK;

	if ( !o_dont_wipe_fn ) {
		utb.actime=utb.modtime=0;
		utime(dirname, &utb);         /* set access and modification time to 0 */
	}

	if ( OK != rmdir(dirname) ) {
		log_message(LOG_ERR, "Could not remove directory '%s': %s\n", dirname, strerror(errno));
		return ERROR;
	}

	log_message(LOG_INFO, "Directory '%s' removed\n", dirname);
	return OK;

}


/*  wipe_filename(char *filename, int delete_file_contents) */
/*  wipe filename by renaming */
/*  return value: */
/*  ERROR   on error */
/*  OK      on success */
int wipe_filename(char *filename, int delete_file_contents)
{
	struct utimbuf utb;
	int ret;

	if ( o_dont_delete ) 
		return OK;

	if ( !o_dont_wipe_fn ) {
		if (delete_file_contents) {
			ret = truncate( filename, 0 );         /* set filesize to 0 */
		}
		utb.actime = utb.modtime = 0;
		utime( filename, &utb );         /* set access and modification time to 0 */
	}
	
	if ( OK != unlink(filename) ) {
		log_message(LOG_ERR, "Could not remove file '%s': %s\n", filename, strerror(errno));
		return ERROR;
	}
	log_message(LOG_INFO, "File '%s' removed\n", filename);
	return OK;
}


/*off_t find_size_by_reading(int fd) */
/*return value: */
/*0       on filure */
/*size    on success */
long long int find_size_by_reading(int fd)
{
	off_t limit, pos, lPos, prev, b_pos = 0;
	char buf[64];
	int i;

	for ( limit=0x7f, i=sizeof(limit); --i; )
		limit = (limit << 8) | 0xff;

	pos = prev = limit;
	while ( pos != b_pos && pos <= limit ) {
		/*        fprintf(stdout,"pos=%20ld   b_pos=%12ld   buf=%x\n",(long int)pos,(long int)b_pos,buf[0]&0xff); */
		
		/*
		//    Version 1.7-7 Aug 07 2008
		//    Treats the lseek error as it cannot set file pointer 
		//	beyond device size instead of sending the error. 
		//	( AIX 5.2 reports here 'Invalid argumment' error )
		//
		//if ( pos != lseek_f(fd,pos,SEEK_SET) )
		//{
		//	fprintf(stderr,"lseek error: %s\n",strerror(errno));
		//	return 0;
		//}
		*/
		
		lPos = lseek_f( fd, pos, SEEK_SET );
		
		if ( ( lPos != pos ) || ( 1 != read(fd, buf, 1) ) ) {
			prev = pos;
			pos = (pos+b_pos) >> 1;
		} else	{
			b_pos = pos;
			pos = (pos+prev) >> 1;
		}
	}

	if ( pos >= limit )
		return 0;

	return lseek_f(fd, 0, SEEK_CUR);
}

long long int get_size_of_device(int fd)
{
	long long size=0;
	int i;
	int hist = 0;

#if defined(BLKGETSIZE)
	long l;
	if ( !ioctl(fd,BLKGETSIZE,&l) )
		size=(long long)l*S_BLKSIZE;

	hist = 1;
#elif defined(DIOC_CAPACITY)
	capacity_type ct;
	if ( !ioctl( fd, DIOC_CAPACITY, &(ct.lba) ) )
		size=(long long)ct.lba * DEV_BSIZE;

	hist = 2;
#elif defined(DIOCGMEDIASIZE)
	off_t media_size;
	if (-1 == ioctl(fd, DIOCGMEDIASIZE, &media_size)) {
		media_size = 0;
	}
	size = media_size;
	hist = 3;
#endif    

	if ( 0 == size || ERROR == size ) {
		size = lseek_f(fd, 0, SEEK_END);
		hist = 4;
	}

	if ( 0 == size || ERROR == size ) {
		size = find_size_by_reading(fd);
		hist = 5;
	}

	if ( 0 == size || ERROR == size ) {
		for (size = 0x7f, i = sizeof(size); --i;  )
			size = (size << 8) | 0xff;
		hist = 6;
	}

	log_message(LOG_INFO, "Device size %lld bytes (%ld kB), method %d\n", size, (long int)(size/1024), hist);

	return size;
}

int wipe_file_slack( char *filename )
{
	char *buff, msg[1024];
	int fd, ret, pass;
	wipe_info_t info;
	struct stat st;
	uid_t uid;
	struct utimbuf time_buf;

	if ( ERROR == lstat(filename, &st) ) {
		log_message(LOG_ERR, "Could not get information on '%s': %s - skipping\n", filename, strerror(errno));
		return ERROR;
	}

	uid = geteuid();
	if (0 != uid && st.st_uid != uid) {
		log_message(LOG_ERR, "You are not owner of '%s' - skipping\n", filename);
		return OK;
	}

	info.size = st.st_size;
	info.slack_size = st.st_size % st.st_blksize ? st.st_blksize - (st.st_size % st.st_blksize) : 0;
	info.st_mode = st.st_mode;
	info.filename = filename;
	info.seed = 0;
	info.prev_seed = 0;
	
	fd = do_open_file_slack(&info);
	if (0 == fd) {
		return OK;
	} else if (ERROR == fd) {
		return ERROR;
	}

	buff = (char *)malloc(st.st_blksize);
	if (NULL == buff) {
		close(fd);
		log_message(LOG_ERR, "Memory allocation error: %s - skipping\n", strerror(errno));
		return OK;
	}

	for ( pass = 0; pass < o_pas_num; pass++ ) {

		info.pass = pass;
		print_pass_name(msg, 1024, &scheme, pass);
		log_message(LOG_INFO, "Wiping slack of '%s' pass %d/%d [%s] started\n", filename, pass+1, o_pas_num, msg);
		
		if (ERROR == lseek_f(fd, st.st_size, SEEK_SET)) {
			log_message(LOG_ERR, "Seek failed on file '%s': %s - skipping\n", filename, strerror(errno));
			ret = ftruncate(fd, st.st_size);
			close(fd);
			free(buff);
			return ERROR;
		}

		info.prev_seed = info.seed;
		info.seed = time(NULL);
		fill_buff(buff, st.st_blksize, pass, 0, info.seed, info.prev_seed);

		if (ERROR == bcwipe_write(fd, (void *)buff, info.slack_size)) {
			log_message(LOG_ERR, "Error writing to '%s' slack: %s\n", filename, strerror(errno));
			ret = ftruncate(fd, st.st_size);
			close(fd);
			free(buff);
			return ERROR;
		}
/*
#ifndef O_SYNC
		(void)fsync(fd);
#endif
*/
		(void)fsync(fd);
		log_message (LOG_INFO, "Wiping slack of '%s' pass %2d/%d completed\n", filename, pass+1, o_pas_num);
	}

	
	ret = ftruncate(fd, st.st_size);

	close(fd);

	free(buff);

	total_size_of_slacks += info.slack_size;
	if ( OK != chmod(filename,st.st_mode) ) {
		log_message(LOG_ERR, "Restore access mask for file '%s': %s\n", filename, strerror(errno));
		return ERROR;
	}

	time_buf.actime  = st.st_atime; 
	time_buf.modtime = st.st_mtime;
	if ( OK != utime(filename,&time_buf) ) {
		log_message(LOG_ERR, "Restore access times of file '%s': %s\n", filename, strerror(errno));
		return ERROR;
	}

	log_message(LOG_INFO, "Wiping slack of '%s' completed\n", filename);

	return OK;
}


/* 
//	wipe_file (char *filename)
//		returns value:
//			ERROR on error
//			OK    on success
*/
int wipe_file( char *filename )
{
	int fd, is_dev, ret;
	struct stat st;
	wipe_info_t info;
	int pass;
	
	if (o_wipe_slacks)
		return wipe_file_slack( filename );

	if ( ERROR == lstat( filename, &st ) ) {
		log_message(LOG_ERR, "Stating file %s: %s - skipping\n", filename, strerror(errno) );
		return ERROR;
	}

	is_dev = is_device(st.st_mode);

	if ( is_dev ){
		if ( ERROR == stat( filename, &st ) ) {
			log_message(LOG_ERR, "Stating file %s: %s - skipping\n", filename, strerror(errno) );
			return ERROR;
		}
		
		if ( !S_ISBLK( st.st_mode ) && !S_ISCHR( st.st_mode ) ) {
			log_message(LOG_ERR, "%s: is not a device - skipping\n", filename );
			return ERROR;	
		}
	}

	memset(&info, 0, sizeof(info));
	info.filename = filename;
	info.size = st.st_size;
	info.slack_size = st.st_size % st.st_blksize ? st.st_blksize - (st.st_size % st.st_blksize) : 0;
	info.st_mode = st.st_mode;

	if (o_interactive && !ask_y_n("Wipe %s (y/[n]/a)?", filename) ) 
		return ERROR;

	if ( 0 == st.st_size && !is_dev ) {
		log_message(LOG_INFO, "Skipping %s '%s': nothing to do\n", get_type_str(st.st_mode), filename);
		return OK;
	}

	log_message(LOG_INFO, "Wiping %s '%s'\n", get_type_str(st.st_mode), filename);

	fd = do_open_file(&info);
	if (ERROR == fd) {
		return ERROR;
	}
	
	if ( is_dev ) {
		info.size = get_size_of_device( fd );
		info.slack_size = 0;
	}

	for ( pass = 0; pass < o_pas_num; pass++ ) {
		info.prev_seed = info.seed;
		info.seed = time( NULL );
		info.pass = pass;
		do_wipe_pass(fd, &info);
	}

	if (o_verify_last_pass) {
		do_verify_pass(fd, &info);
	}
	
	if ( !o_dont_delete )
		ret = ftruncate(fd, 0);

	close(fd);

	return OK;
}

int wipe_file_by_type(char *filename, char *dirname) 
{
	struct stat buff;
	int	result;
	
	if ( OK != lstat(filename, &buff) )	{
		log_message(LOG_ERR, "Cannot stat '%s': %s - skipping\n", filename, strerror(errno));
		return ERROR;
	}

	if ( S_ISDIR(buff.st_mode) ) {
		if (dirname) {
			if (0 == strcmp(dirname, ".") || 0 == strcmp(dirname, ".."))
				return OK;
		} else if (!o_recurse) {
			log_message(LOG_ERR,"'%s' is a directory - skipping\n", filename);
			return OK;
		}
		result = wipe_dir(filename);
	} else if ( S_ISREG(buff.st_mode) ) {
		result = wipe_file( filename );
		if ( OK == result ) {
			result = wipe_filename(filename, 1);
		}
	} else if ( S_ISCHR(buff.st_mode) || S_ISBLK(buff.st_mode) || S_ISLNK(buff.st_mode) )  {
		if ( o_wipe_dev ) {
			result = wipe_file(filename);
		} else {
			result = wipe_filename(filename, 0);
		}
	} else if ( S_ISFIFO(buff.st_mode) || S_ISSOCK(buff.st_mode) ) {
		result = wipe_filename(filename, 0);
	} else {
		log_message(LOG_ERR, "'%s': unsupported file type - skipping\n", filename);
		return ERROR;
	}
	return result;
}

int wipe_dir( char *dirname )
{
	struct dirent *d;
	DIR *fd;
	char fn[NAME_MAX+PATH_MAX];

	if (NULL == dirname) 
		return ERROR;

	if (o_interactive && !ask_y_n("Enter directory %s (y/[n]/a)?", dirname))
		return ERROR;

	/* check directory access */
	if ( !o_wipe_slacks && OK != access(dirname,X_OK|R_OK|W_OK)) {
		if ( !o_force )	{
			log_message(LOG_ERR, "Could not enter directory '%s': %s - skipping\n", dirname, strerror(errno));
			return ERROR;
		}
		if ( OK != chmod(dirname,S_IRUSR|S_IWUSR|S_IXUSR) ) {
			log_message(LOG_ERR, "Could not chmod directory '%s': %s - skipping\n", dirname, strerror(errno));
			return ERROR;
		}
	}

	fd = opendir(dirname);
	if ( NULL == fd ) {
		log_message(LOG_ERR, "Could not open directory '%s': %s - skipping\n", dirname, strerror(errno));
		return ERROR;
	}

	log_message(LOG_INFO, "Entering directory '%s'\n", dirname);

	while ( NULL != (d = readdir(fd)) ) {
		snprintf(fn, NAME_MAX+PATH_MAX, "%s/%s", dirname, d->d_name);
		wipe_file_by_type(fn, d->d_name);
	}

	closedir(fd);

	if ( o_dont_delete )
		return OK;

	if (o_interactive && !ask_y_n("Remove directory '%s' (y/[n]/a)?", dirname) )
		return ERROR;

	if ( OK != rmdir(dirname) ) {
		log_message (LOG_ERR, "Error while removing '%s': %s\n",dirname,strerror(errno));
		return ERROR;
	}
	log_message(LOG_INFO, "Directory '%s' removed\n", dirname);

	return OK;
}



/* -------------- wipe free space routines -------------- */
char *get_mount_point(char *path)
{
	return path;
}

long long int get_free_space(char *path)
{
	int res;
	struct statvfs_f sfs;
	res = statvfs_f( path, &sfs );

	if (res != 0) {
		log_message(LOG_ERR, "statvfs(%s): %s\n",path,strerror(errno));
		return 0;
	}

	return (off_t)sfs.f_bfree * sfs.f_frsize;
}

int can_wipe_all_free_space(char *path)
{
	struct statvfs_f sfs;
	int res;
	
	res = statvfs_f( path, &sfs );
	if (res != 0) {
		log_message(LOG_ERR, "Cannot get free space via statvfs(%s): %s\n", path, strerror(errno));
		return 0;
	}
	return (sfs.f_bfree == sfs.f_bavail || geteuid() == 0);
}

static wipe_info_t *fs_alloc_wipe_info( wipe_info_t **pphead, char *filename, int pass )
{
	wipe_info_t *ptr;
	
	ptr = (wipe_info_t *)malloc( sizeof( wipe_info_t ) );
	if ( !ptr )
		return NULL;
	
	memset(ptr, 0, sizeof(wipe_info_t));
	ptr->filename = strdup(filename);
	if (!ptr->filename) {
		free(ptr);
		return NULL;
	}
	ptr->prev_seed = ptr->seed;
	ptr->seed = time(NULL);
	ptr->pass = pass;
	ptr->next = *pphead;
	*pphead = ptr;

	return ptr;
}

#define MAX_FILE_SIZE   1024*1024*1024

int fs_create_files( wipe_info_t **head, char *path, int pass, long long int *ret_size )
{
	char buff[BUFFSIZE];
	char tmp[PATH_MAX+NAME_MAX];
	int fd, res, j, write_fail;
	long long int file_size, ss, free_space, x;
	time_t  start_time;
	int cnt, err;
	seed_t  seed, prev_seed;


	/*
	//	The is a strange report about HPUX 11.23.
	//	We are running it on HPUX 11.23 with the option vIF and what happens is
	//	that a directory is created with a 5 digit random number appended and in
	//	it is created up to 26 1GB files, each one prefixed by a letter a-z.
	//	Once z is reached it stops, removes the files and then starts on the
	//	next pass.
	//
	//	Version 1.7-3
	//	The 'cnt' is a name of the temporary file and extention is random.
	*/
	cnt = 0;
	start_time = time(NULL);
	write_fail = 0;
	file_size  = 0;
	free_space = get_free_space(path);

	seed = prev_seed = 0;
	print_pass_name(buff, BUFFSIZE, &scheme, pass);
	log_message(LOG_INFO, "Free space wiping pass %2d/%-2d [%s] started. Total free space %lld kB\n", pass+1, o_pas_num, buff, free_space >> 10);

	if ( 0 == free_space ) {
		if (1 != pass) {
			log_message(LOG_ERR, "No free space available on this filesystem\n");
			return ERROR;
		}
		return OK;
	}
	
	
	while ( TRUE ) {
		ss = 0;

		snprintf( tmp, PATH_MAX+NAME_MAX, "%s/%03x.XXXXXX", path, cnt++ );
		fd = mkstemp( tmp );
		if ( ERROR == fd ) {
			if ( ENOSPC == errno ) {
				res = 0;
				break;
			} 
			log_message(LOG_ERR, "mkstemp(%s): %s\n", tmp, strerror( errno ) );
			goto error_out; 
		}

		if ( !fs_alloc_wipe_info( head, tmp, pass ) ) {
			close( fd );
			log_message(LOG_ERR, "Error in allocating memory\n" );
			goto error_out; 
		}

		free_space = get_free_space(path);
		do {
			
			if (free_space > BUFFSIZE) {
				j = BUFFSIZE;
			} else {
				j = 1;
				x = free_space;
				while (x >> j) {
					j++;
				}
				j = 1 << (j-1);
				if (write_fail) {
					j = (j+1) / 2;
					write_fail = 0;
				}
			}
			prev_seed = seed;
			seed = time(NULL);	
			fill_buff( buff, j, pass, ss, seed, prev_seed );
			res = bcwipe_write( fd, buff, j );
			if ( ERROR != res ) {
				file_size += res;
				ss += res;
				if (ret_size) {
					*ret_size = file_size;
				}
			}
			
			free_space = get_free_space(path);

			if ( o_verbose ) {
				print_wiping_progress(free_space+file_size, file_size, start_time, pass);
			}

			if ( ERROR == res ) {
				if (ENOSPC == errno) {
				    res = ENOSPC;
				}
				break;
			}
			
			if ( MAX_FILE_SIZE <= ss ) {
				res = 0;
				break;
			}

		} while ( 1 || ERROR != res );

		write_fail = (0 == ss);

		(void)fsync(fd);

		close( fd );
	}
	
	if (OK == res) {
		if (o_verbose) {
			fprintf(stdout, "\n");
		}
		log_message(LOG_INFO, "Free space wiping pass %2d/%-2d completed. Processed %lld kB, remaining %lld kB\n", 
		pass+1, o_pas_num, file_size >> 10, get_free_space(path) >> 10);
		return OK;
	}
	
error_out:
	if (o_verbose) {
		fprintf(stdout, "\n");
	}
	log_message(LOG_ERR, "Free space wiping pass %2d/%-2d failed. Processed %lld kB\n", pass+1, o_pas_num, file_size >> 10);

	return ERROR;
}

static int fs_verify_files( wipe_info_t *head, long long int size, int pass )
{
	char buff[BUFFSIZE];
	char verifyBuff[BUFFSIZE];
	time_t start_time;

	int fd;
	long long int s, ss;
	ssize_t res, x;
	wipe_info_t *ptr;

	log_message(LOG_INFO, "Free space wiping pass %2d verification started\n", pass+1 );
	
	s = 0;
	start_time = time(NULL);

	for (ptr = head; ptr; ptr = ptr->next ) {

		fd = open( ptr->filename, O_RDONLY );
		if ( ERROR == fd ) {
			log_message(LOG_ERR,"Error in opening file '%s' for verification: %s\n", ptr->filename, strerror(errno) );
			goto error_out;
		}
		
		ss = 0;
		do {
			res = bcwipe_read( fd, buff, BUFFSIZE );
			if ( ERROR == res ) {
				log_message(LOG_ERR, "Error in reading file '%s' for verification: %s\n", ptr->filename, strerror(errno) );
				close( fd );
				goto error_out;
			}

			fill_buff( verifyBuff, res, ptr->pass, ss, ptr->seed, ptr->prev_seed );
			x = verify_buffer(buff, verifyBuff, res);
			if ( 0 != x ) {
				log_message( LOG_ERR, "Verification error. File '%s', offset %lld\n"
						      "Expected %02Xh, read %02Xh\n", ptr->filename, s+x-1, (unsigned char)verifyBuff[x-1], (unsigned char)buff[x-1]);
				close( fd );
				goto error_out;
			}

			s  += res;
			ss += res;

			if ( o_verbose ) {
				print_verification_progress(size, s, start_time, pass);
			}
		} while ( res > 0 );

		close( fd );
	}

	if (o_verbose) {
		fprintf(stdout, "\n");
	}
	log_message(LOG_INFO, "Free space wiping pass %2d verified %lld kB\n", pass+1, s >> 10);
	return OK;
error_out:

	if (o_verbose) {
		fprintf(stdout, "\n");
	}
	log_message(LOG_ERR, "Free space wiping pass %2d verification failed\n", pass+1);
	return ERROR;

}

int allocate_free_space( char *path, int pass, int verify )
{
	wipe_info_t *fs_head = NULL;
	long long int size;
	int res;

	res = fs_create_files( &fs_head, path, pass, &size );
	if (OK == res) {
		if ( o_nas_delay ) {
			if ( o_verbose ) {
				fprintf( stdout, "Sleeping %d seconds\n", o_nas_delay );
			}
			sleep(o_nas_delay);
		}

		if ( verify ) {
			res = fs_verify_files( fs_head, size, pass );
		}
	}
	free_list( fs_head );
	return res;
}

int delete_temp_dir(char *dirname)
{
	struct dirent *d;
	DIR *fd;
	struct stat buff;
	char fn[NAME_MAX+PATH_MAX];

	if ( NULL == dirname ) {
		return ERROR;
	}
	
	fd = opendir(dirname);
	if ( NULL == fd ) {
		log_message(LOG_ERR, "opendir(%s): %s \n",dirname,strerror(errno));
		return ERROR;
	}

	while ( NULL != (d=readdir(fd)) ) {
		snprintf(fn, NAME_MAX+PATH_MAX, "%s/%s", dirname, d->d_name);
		if ( OK != lstat(fn,&buff) ) {
			log_message(LOG_ERR,"lstat(%s): %s\n",fn,strerror(errno));
			continue;
		}
		if ( S_ISDIR(buff.st_mode) ) {    /*directory */
			if ( 0 == strcmp(d->d_name,".") || 0 == strcmp(d->d_name,"..") )
				continue;
			delete_temp_dir(fn);
		} else {
			if ( 0 != unlink(fn) ) {
				log_message(LOG_ERR, "unlink(%s): %s\n", fn, strerror(errno));
				return ERROR;
			}
		}
	}
	closedir(fd);

	if ( OK != rmdir(dirname) ) {
		log_message(LOG_ERR, "rmdir(%s): %s\n", dirname, strerror(errno));
		return ERROR;
	}
	sync();
	return OK;
}

char *make_temp_dir( char *path, char *result, int result_sz )
{
	int fd;
	/*	GCC reports warning 'mktemp is dangerous, better use mkstemp' */
	snprintf( result, result_sz, "%s/bcwipe-wiping_free_space-XXXXXX", path );
	fd = mkstemp( result );
	if ( ERROR == fd )	{
		log_message(LOG_ERR, "Failed to create temporary file via mkstemp(%s): %s\n", result, strerror(errno));
		return NULL;
	}
	close(fd);
	unlink( result );

	if (OK != mkdir(result, 0700)) {
		log_message(LOG_ERR ,"Failed to create a directory via mkdir(%s): %s\n",result,strerror(errno));
		return NULL;
	}
	return result;
}

/*
//	wipe_free_space
//	1. Creates new files and fills tmen by wiping data for all free space available.
//	All creates file names are stored in list 
//	2. The last pass free space allocation verifies the new files by the list.
//	3. Removes the new directory.
*/
int wipe_free_space(char *path)
{
	int i, res;
	char tmp_dir[PATH_MAX];

	if (o_interactive && !ask_y_n("Wipe free space in %s (y/[n])?",  get_mount_point(path)))
		return OK;

	if (OK != access(path, W_OK|R_OK|X_OK)) {
		log_message(LOG_ERR, "Cannot wipe free space in %s: %s\n", path, strerror(errno));
		return ERROR;
	}

	if (!can_wipe_all_free_space(path) && 
		o_interactive &&
		!ask_y_n("You can not wipe ALL free space on %s.\nContinue (y/[n])?", get_mount_point(path)))
			return OK;
	
	log_message(LOG_INFO, "Wiping free space on %s\n", get_mount_point( path ) );

	for (i = 0; i < o_pas_num; i++) {
	
		if (!make_temp_dir(path, tmp_dir, PATH_MAX)) {
			log_message(LOG_ERR, "Could not create temporary directory for %s\n", get_mount_point(path));
			return ERROR;
		}

		res = allocate_free_space( tmp_dir, i, o_verify_last_pass && (i == o_pas_num-1) );
		if (OK != res) {
			return res;
		}

		res = delete_temp_dir(tmp_dir);
		if (OK != res) {
			log_message(LOG_ERR, "Error deleting temporary files under %s\n", get_mount_point(path));
			return res;
		}
	}
	log_message(LOG_INFO, "Wiping free space on '%s' completed\n", get_mount_point( path ) );
	
	return OK;
}

static void exit_handler(void) 
{
	cleanup_scheme(&scheme);
	close_log();
}

static void signal_handler(int sig) 
{
	if (!ask_y_n("\nDo you want to stop wiping (y/[n])?")) {
	    return;
	}
	log_message(LOG_ERR, "Process was interrupted by user\n");
	exit_handler();
	exit(-1);
}

/* -------------- end of wipe free space routines -------------- */


int main(int argc, char* argv[])
{
	int i, result;

	/* turn off buffering*/
	setvbuf(stderr,(char*)NULL,_IONBF,0);
	setvbuf(stdout,(char*)NULL,_IONBF,0);
	
	signal(SIGINT,  signal_handler);
	signal(SIGQUIT, signal_handler);
	signal(SIGABRT, signal_handler);
	signal(SIGTERM, signal_handler);
	atexit(exit_handler);

	result = parseOptions( &argc, &argv );
	if (0 != result) {
		return result;
	}

	log_message(LOG_FILE, "**** BCWipe %s started ****\n", VERSION);
	result = init_scheme(&scheme);
	if (OK != result) {
		return ERROR;
	}
	
	result = StartRandom();
	if (OK != result) {
		return result;
	}
	log_message(LOG_INFO, "Wiping scheme: %s, %d pass(es)\n", scheme.name, scheme.num_passes);
	
	if (o_wipe_free_space) {
		result = wipe_free_space(argv[0]);
		if (OK == result && o_wipe_slacks) {
			o_recurse = TRUE; /* free space wiping complete, now wipe file slacks        */

		} else {
			StopRandom();
			return result;
		}
	}

	if (o_nas_wiping) {
		result = nas_wiping(argc, argv);
	} else {
		for ( i = argc; i; i-- ) {
			wipe_file_by_type(argv[i-1], NULL);
		}
	}
	if (o_wipe_slacks) {
		if (total_size_of_slacks > 4096) {
			log_message(LOG_INFO, "Wiped %d kB of file slacks\n",(unsigned int)(total_size_of_slacks/1024));
		} else {
			log_message(LOG_INFO, "Wiped %d bytes of file slacks\n",(unsigned int)(total_size_of_slacks));
		}
	}

	StopRandom();

	return result;
}


